
/********** global parameters **************************/

// change this to define if context ( November 1998 ) is to be used
#define USE_CONTEXT
//#undef USE_CONTEXT

// change this to define if new ( November 1998 ) rama map is to be used
#define USE_NEW_RAMA_MAP
//#undef USE_NEW_RAMA_MAP

#define M 20
#ifdef USE_NEW_RAMA_MAP
#define MRAMA 11
#else
#define MRAMA 4
#endif

#define MDSSP 6
#define MCTXT 10
#define Tmax 5000
#define NCEN 9
#define ALPH_1 0.5
#define ALPH_2 0.2


// #define NCOUNT 1.0
// #define EPS 0.001

/****************** structures ******************************/
  
  

struct model{ 
  int   N ;                 // Number of nodes
  float bground[M] ;        // Background amino acid frequencies
  float rground[MRAMA] ;    // Background rama-type frequencies
  float dground[MDSSP] ;    // Background DSSP labelfrequencies
  float *prior ;            // Prior probabilities
  float **dih ;             // Dihedral angles of nodes 
  char  *ss ;               // HSSP structure label of nodes 
  float **a ;               // Transition matrix 
  float **b ;               // Emission probabilities for amino acids
  float **r ;               // Emission probabilities for Ramachandran regions
  float **d ;               // Emission probabilities for DSSP assignments
#ifdef USE_CONTEXT
  float cground[MCTXT] ;    // Background context frequencies
  float **c ;               // Emission probabilities for context assignments
#endif
  double **logb ;            // logarithm of b contribution to probability, with EPS, NCOUNT
} ;

struct parent { 
  char  code[4] ;           // PDB code
  char  chainID ;           // chain ID
  short int T ;            // Length of parent
  char  *O_ptr ;            // Pointer to residue sequence of parent
  char  *O_ss;              // Pointer to HSSP structure label
  char  *O_rama;            // Pointer to HSSP structure label
  float *O_phi; 
  float *O_psi; 
  float *O_omg;            // Pointers to observed dihedral angles 
  float **O_profile;
  float *O_gap;
  float *O_ins;
  char  domainID;
  char  *O_context;
  float *weight;
  float *x;         // C-alpha coordinates
  float *y;         // C-alpha coordinates
  float *z;         // C-alpha coordinates
  float *bestcf;
  float *qscore;
  float *cscore;
  short int *tf;
  short int *cen;
  short int *resseq;
}; 

struct dbaserec { 
  short int seq_list;     // 
  short int resSeq ;      // PDB file residue number
  char code[4];           // PDB code
  char chainID;           // Chain letter
  char seq_resid;         // Residue
  char ss;                // DSSP assignment: H,G,E,S,T, or _
  char rotamer; 
  float xyzca[3];         // C-alpha coordinates
  float dih[3];           // Diheral angles 
  float csum; 
  float profile[20];      // Profile
  float gap_freq;         // Gap frequency
  float ins_freq;         // Insertion frequency
  char  domainID;         // Domain ID 
  char  context;          // Context
  short int centroid;   // Blank
  short int generic[2];   // Blank
} ; 


struct cluster{
  // Cluster summary information
  int ID;
  int L;
  int Lwi;
  int ohang;
  // Wi[L+2*ohang][20]
  float **Wi;
  float **Pr;
  // parent paradigm;
  // need to add confidence curve (score --> confidence)
  float *phi, *psi, *omg;
  float mdacut, dmecut ;
  int Lcf;
  float cfscore[20], cf[20] ;
  // parameters for converting to cf
  float b0,b1,b2,ww ;

  // Member information
  int Nmembers;
  parent *member;
  float *score;
  char *tf;
  // cluster paradigm
  char pcode[5];
  char pchain;
  int pstart;
} ;







