// change this to undef if ithree is not used in main.
//#undef USE_ITHREE
#define USE_ITHREE

// change this to def if new ( November 1998 ) rama map is to be used
#define USE_NEW_RAMA_MAP
//#undef USE_NEW_RAMA_MAP

// change this to def if context ( November 1998 ) is to be used
#define USE_CONTEXT
//#undef USE_CONTEXT

#include <stdio.h> 
#include <iostream.h> 
#include <fstream.h>
#include <string.h>
#include <stdlib.h>
#include <iostream.h>
#include <iomanip.h> 
#include <math.h>

#include "hmm.h"
#include "ss_maps.h"
#include "hmm_model_tools.h"

// flagrat_brdc_fast    uses bprof, which calculates the multinomial score
// pseudocounts included in r,d,c 28-Jan-98
// now uses bprof_bg with pseudocounts  6-OCT-98 

#ifdef USE_CONTEXT
double flagrat_brdc_fast ( int iflag ,
                int m, 
	        int t, 
	        int *O, 
	        struct parent pa, 
	        struct model mod  ){

  int k, jflag ; 
  double brat, rrat, drat, crat ; 
  brat = 1.; rrat = 1.; drat = 1.; crat=1.;
  
  // Context symbol

  jflag = iflag / 8 ;
  if ( jflag > 0 ){
    k = icontext ( pa.O_context[t] ) ;
    if ( k > -1 ){
      crat = ( mod.c[m][k] + EPS * mod.cground[k] )/((1+EPS)*mod.cground[k]) ;
    }
  }

  // Three state DSSP symbol 
  iflag = iflag % 8 ;
  jflag = iflag / 4 ; 
  if (jflag > 0) {
    k = threestate( pa.O_ss[t] ) ;
    if ( k > -1 ) {
     drat = (mod.d[m][k] + EPS * mod.dground[k] )/((1+EPS)*mod.dground[k]) ; 
    }
  }

  // Ramachandran region
  iflag = iflag % 4 ;
  jflag = iflag / 2 ; 
  if (jflag > 0) {
    k = irama( pa.O_rama[t] ) ;
    if ( k > -1 ) {
     rrat = (mod.r[m][k] + EPS * mod.rground[k] )/((1+EPS)*mod.rground[k]) ; 
    }
  }
  
  // Amino acid emission
  jflag = iflag % 2 ;
  if (jflag > 0) {
    brat = bprof_fast(m,t,pa,mod) ; // default, for profiles
    //k=O[t];brat = (mod.b[m][k] + EPS * mod.bground[k] )/((1+EPS)*mod.bground[k]) ; //for_single_sequences use this line to replace bprof_fast
  }

  
  return brat * rrat * drat * crat; 
  
}

#endif

/******************************************************************/

double bprof_fast  ( int m, 
	        int t, 
	        struct parent pa, 
	        struct model mod  ){

  int a; 
  double logsum=0.;

  // Amino acid emission

  //  now uses pseudocounts based on bground  6-OCT-98
  for ( a=0 ; a<M ; a++ ) {
    if ( pa.O_profile[t][a] != 0 ) {
      logsum += pa.O_profile[t][a] * mod.logb[m][a] ;
    }
  }

  return exp(logsum) ;

}


//            
//				    November 9, 1998
//
//   modelf.C - read model from file
//		
//              Input:   filename
//              Output:  Pointer to model structure containing
//                       N       - #nodes including unknown
//                       bground - background frequencies 
//                       prior   - priors for nodes
//                       a       - transition probabilities
//                       b       - emission probabilities
//
//
// keywords in hmm files
#define BGFREQS "default_bg_freqs"
#define NODE "node"
#define NODE_SS "node_ss"
#define NODE_RAMA "node_rama"
#define NODE_CONTEXT "node_context"
#define TRANSIT_FREQS "transit_freqs"
#define NUM_NODES "num_nodes"
#define UNK_NODE "unk_node"

void modelf ( char file[], struct model *modl ) 
{

  char word[200];
  int i, j, n, N, nnodes;  
  float sumd[MDSSP], sumr[MRAMA], sumc[MCTXT];
  float x;
  extern int NOT_STATE_FLAG; 

  ifstream in(file);
  if ( in.bad() ) {
    cerr << "Unable to open model file\n";
    exit (8);
  }


#ifdef USE_ITHREE
  // default DSSP vector order :  HEGST_    i.e. ss_map 6 HEGST_
  ithree[0] = threestate('H') ;
  ithree[1] = threestate('E') ;
  ithree[2] = threestate('G') ;
  ithree[3] = threestate('S') ;
  ithree[4] = threestate('T') ;
  ithree[5] = threestate('_') ;
#endif

  // Background frequencies
  cout << "Getting " << BGFREQS << endl << flush ;

  in >> word;
  while (!in.eof() && strcmp(word,BGFREQS)!=0) {
    in.ignore(100,'\n');
    in >> word ;  

#ifdef USE_ITHREE
    // DSSP vector order :  i.e: ss_map 6 HEGST_
    if (strcmp(word,"ss_map")==0) {
      in >> word ; sscanf(word,"%d",&j) ;
      in >> word ;
      cout << "Three-state assigments: " << j << "   " ;
      for (i=0; i<j ; i++) {
        ithree[i] = threestate(*(word+i)) ;
        cout << *(word+i) << "=" << ithree[i] << " " ;
      }
      cout << endl << flush ;
      in.ignore(100,'\n');
      in >> word ;  
    }
#endif
  }

  for (i=0;i<MRAMA;i++) sumr[i]=0.;
  for (i=0;i<MDSSP;i++) sumd[i]=0.;
#ifdef USE_CONTEXT
  for (i=0;i<MCTXT;i++) sumc[i]=0.;
#endif 

  if (in.eof()) {
    in.seekg(0,ios::beg);
    cout << "KEYWORD default_bg_freqs NOT FOUND!" << endl << flush;
    cout << "using 0.05 for all backgroud freqs" << endl << flush;
    for ( i = 0 ; i < M ; i++ ) modl->bground[i] = 0.05 ;
  } else {
    for ( i = 0 ; i < M ; i++ ) in >> modl->bground[i] ;
  }
  cout << "Background AA freqs: " ;
  for (i = 0 ; i < M ; i++ ) cout << setw(8) << modl->bground[i] << " " ;
  cout << endl << flush ;

  // Number of nodes

  cout << "Getting " << NUM_NODES << endl << flush ;
  in >> word;
  while(!in.eof() && strcmp(word,NUM_NODES)!=0) {
    in.ignore(100,'\n');
    in >> word ;  
  }
  if (in.eof()) {
    in.seekg(0,ios::beg);
    cout << "KEYWORD num_nodes NOT FOUND!" << endl << flush;
    exit(8);
  }
  in >> nnodes ;
  modl->N = nnodes + 1 ;
  N  = modl->N ;  

  // Memory allocation 
  cout << "allocating " << N << " nodes." << endl << flush ;

  modl->prior = new float[N] ;
  modl->dih   = new float*[N] ; 
  for ( i = 0 ; i < N ; i++ ) modl->dih[i]   = new float[3] ; 
  modl->ss = new char[N] ;
  modl->a     = new float*[N] ; 
  for ( i = 0 ; i < N ; i++ ) modl->a[i]     = new float[N] ; 
  modl->b       = new float*[N] ;
  for ( i = 0 ; i < N ; i++ ) modl->b[i]     = new float[M] ; 
  modl->r       = new float*[N] ;
  for ( i = 0 ; i < N ; i++ ) modl->r[i]     = new float[MRAMA] ; 
  modl->d       = new float*[N] ;
  for ( i = 0 ; i < N ; i++ ) modl->d[i]     = new float[MDSSP] ; 
#ifdef USE_CONTEXT
  modl->c       = new float*[N] ;
  for ( i = 0 ; i < N ; i++ ) modl->c[i]     = new float[MCTXT] ; 
#endif
  modl->logb       = new double*[N] ;
  for ( i = 0 ; i < N ; i++ ) modl->logb[i]     = new double[M] ; 

  // Unknown node 

  cout << "Getting " << UNK_NODE << endl << flush ;
  while(!in.eof() && strcmp(word,UNK_NODE)!=0) {
    in.ignore(100,'\n');
    in >> word ;
  }
  if (in.eof()) {
    in.seekg(0,ios::beg);
    cout << "KEYWORD unk_node NOT FOUND!" << endl << flush;
    cout << "using 0.01 for prior of unk_node" << endl << flush;
    cout << "using 0.05 for freqs of unk_node" << endl << flush;
    modl->prior[0] = 0.01;
    for ( i = 0;  i < M ; i++ ) modl->b[0][i] = 0.05;    
  } else {
    in >> word ;
    in >> modl->prior[0];
    for ( i = 0;  i < M ; i++ ) in >> modl->b[0][i];    
  }
  modl->dih[0][0] = -75. ;
  modl->dih[0][1] = -15. ;
  modl->dih[0][2] = 180. ;
  modl->ss[0] = '_' ;

  // assign global NOT_STATE_FLAG
  if (  modl->prior[0] < 1e-8 ) {
    NOT_STATE_FLAG = 0 ; // NOT-STATE not used in the model
  } else {
    NOT_STATE_FLAG = 1 ; // NOT-STATE in use for model
  } 

  // get DSSP distribution for unknown state
  while(!in.eof() && strcmp(word,"unk_node_ss")!=0) {
    if (strcmp(word,NODE)==0) break ;
    in.ignore(100,'\n');
    in >> word ;
  }
  if (strcmp(word,"unk_node_ss")==0) {
    cout << "Found unk_node_ss line." << endl << flush ;
    in >> modl->ss[0] ;
    for ( i = 0;  i < MDSSP ; i++ ) {
      in >> modl->d[0][i];    
      sumd[i] += modl->prior[0] * modl->d[0][i] ;
    }
  }

  // get RAMA distribution for unknown state
  while(!in.eof() && strcmp(word,"unk_node_rama")!=0) {
    if (strcmp(word,NODE)==0) break ;
    in.ignore(100,'\n');
    in >> word ;
  }
  if (strcmp(word,NODE)!=0) {
    cout << "Found unk_node_rama line." << endl << flush ;
    in >> word ;  // RAMA region letter
    for ( i = 0;  i < MRAMA ; i++ ) {
      in >> modl->r[0][i];    
      sumr[i] += modl->prior[0] * modl->r[0][i] ;
    }
  }
  if (in.eof()) {
    cout << "MISSING line(s) in model file before nodes!! " << endl ;
    exit(8);
  }

  // Known nodes 

  cout << "Getting " << NODE << " lines." << endl ;
  for ( n = 1;  n < N ; n++ ){

    while(!in.eof() && strcmp(word,NODE)!=0) {
      if (strcmp(word,TRANSIT_FREQS)==0) break ;
      in.ignore(100,'\n');
      in >> word ;
    }
    if (strcmp(word,TRANSIT_FREQS)==0) break ;
    if (in.eof()) {
      cout << "MISSING node line(s) in model file!! " << n << endl ;
      exit(8);
    }
    if (strcmp(word,NODE)==0) {
      in >> word;   // node number
      in >> word ;  // node character
      in >> word ;  // motif number
      in >> modl->prior[n] ;
      in >> modl->dih[n][0] ;
      in >> modl->dih[n][1] ;    
      in >> modl->dih[n][2] ;
      in >> word ;  // component= field
      in >> word ;  // component number
      in >> word ;  // class= field
      in >> word ;  // ss= field
      modl->ss[n] = word[3];  // overwritten if node_ss lines are present.
      in.ignore(100,'\n');
      for ( i = 0;  i < M ; i++ ) in >> modl->b[n][i];   
    }

    // get DSSP distribution for node
    while(!in.eof() && strcmp(word,NODE_SS)!=0) {
      if (strcmp(word,NODE)==0) break ;
      if (strcmp(word,TRANSIT_FREQS)==0) break ;
      in.ignore(100,'\n');
      in >> word ;
    }
    if (strcmp(word,NODE_SS)==0) {
      in >> word ;  // node number
      in >> word ;  // node letter
      in >> modl->ss[n] ;  // predimonant DSSP label
      for ( i = 0;  i < MDSSP ; i++ ) {
        in >> modl->d[n][i];    
        sumd[i] += modl->prior[n] * modl->d[n][i] ;
      }
    }

    // get RAMA , Ramachandran angle region emission vector
    while(!in.eof() && strcmp(word,NODE_RAMA)!=0) {
      if (strcmp(word,NODE)==0) break ;
      if (strcmp(word,TRANSIT_FREQS)==0) break ;
      in.ignore(100,'\n');
      in >> word ;
    }
    if (strcmp(word,NODE_RAMA)==0) {
      in >> word ;  // node number
      in >> word ;  // node letter
      in >> word ;  // rama letter
      for ( i = 0;  i < MRAMA ; i++ ) {
        in >> modl->r[n][i];    
        sumr[i] += modl->prior[n] * modl->r[n][i] ;
      }
    }

#ifdef USE_CONTEXT
    // get CONTEXT, context emission vector
    while(!in.eof() && strcmp(word,NODE_CONTEXT)!=0) {
      if (strcmp(word,NODE)==0) break ;
      if (strcmp(word,TRANSIT_FREQS)==0) break ;
      in.ignore(100,'\n');
      in >> word ;
    }
    if (strcmp(word,NODE_CONTEXT)==0) {
      in >> word ;  // node number
      in >> word ;  // node letter
      in >> word ;  // context letter
      for ( i = 0;  i < MCTXT ; i++ ) {
        in >> modl->c[n][i];    
        sumc[i] += modl->prior[n] * modl->c[n][i] ;
      }
    }
#endif

  }

  // save background frequencies for RAMA labels
  x = 0.; for (i=0;i<MRAMA;i++) x += sumr[i];
  if (x > 0.) {
    for ( i = 0;  i < MRAMA ; i++ ) modl->rground[i] = sumr[i]/x ;
  } else {  // this should never be used!
    for ( i = 0;  i < MRAMA ; i++ ) modl->rground[i] = 1./((float) MRAMA) ;
  }
  cout << "Background RAMA labels: "  << setiosflags(ios::fixed ) ;
  for ( i = 0;  i < MRAMA ; i++ ) {
    cout << setw(8) << setprecision(4) << modl->rground[i];
  }
  cout << endl << flush ;


  // save background frequencies for DSSP labels
  x = 0.; for (i=0;i<MDSSP;i++) x += sumd[i];
  if (x > 0.) {
    for ( i = 0;  i < MDSSP ; i++ ) modl->dground[i] = sumd[i] ;
  } else {  // this should never be used!
    for ( i = 0;  i < MDSSP ; i++ ) modl->dground[i] = 1./((float) MDSSP) ;
  }
  cout << "Background DSSP labels: " ;
  for ( i = 0;  i < MDSSP ; i++ ) {
    cout << setw(8) << setprecision(4) << modl->dground[i];
  }
  cout << endl << flush ;

#ifdef USE_CONTEXT
  // save background frequencies for CONTEXT labels
  x = 0.; for (i=0;i<MCTXT;i++) x += sumc[i];
  if (x > 0.) {
    for ( i = 0;  i < MCTXT ; i++ ) modl->cground[i] = sumc[i] ;
  } else {  // this should never be used!
    for ( i = 0;  i < MCTXT ; i++ ) modl->cground[i] = 1./((float) MCTXT) ;
  }
  cout << "Background CTXT labels: " ;
  for ( i = 0;  i < MCTXT ; i++ ) {
    cout << setw(8) << setprecision(4) << modl->cground[i];
  }
  cout << endl << flush ;
#endif 

  // Transition matrix 

  cout << "Getting " << TRANSIT_FREQS << endl << flush ;
  while(!in.eof() && strcmp(word,TRANSIT_FREQS)!=0 ) {
    in.ignore(100,'\n');
    in >> word ;
  }
  if (in.eof()) {
    cout << "MISSING transition freqs in model file!! " << endl << flush;
    exit(8);
  }

  // read transition frequencies
  if (strcmp(word,TRANSIT_FREQS)==0 ) {
    in.ignore(300,'\n');
    for ( i = 0; i < modl->N; i++ ){
      for ( j = 0; j < modl->N; j++ ) in >> modl->a[i][j] ;
    } 
  } 
  cout << "Done reading " << N*N << " transition freqs" << endl << flush ;

  in.close();

}


//
//                                             April 16, 1998
//
//     write_model.C - Write model to output file  
//
//            Input:  1. Model structure 
//                    2. Initial model-file name 
//                    3. Output model-file name 
//                    4. 0 if no new nodes
//                       -1 if L unit was added to model 
//                       # of original node if 
//                       new node was created by splitting
//                       ( structural information copied to final (new) node )
//                    5. Header string 
//

void write_model (  struct model mod, 
		    char infile[], 
		    char outfile[] ,  
		    int spnode, 
		    char header[] ) {
  
  int i, j, jj, N ; 
  char line[300] ;
  char *l_ptr ;
  char spline[300], spline_ss[300], spline_rama[300], spline_context[300]; 
#ifdef USE_NEW_RAMA_MAP
  char ramach[] = "HGBEdbeLlxc" ;
  extern float ramacen[MRAMA][2] ;
#else
  char ramach[]="HELX";
#endif
  char context_ch[] = "amnhdstul_";
  char dsspletter[]="HEGST_";
  float x;

  N = mod.N ;

  ifstream in ( infile ) ; 
  ofstream out( outfile ) ;

  out << header << endl ; 

  // Print number of nodes 

  // Get num_nodes line ***required***
  in.getline( line, sizeof(line) ) ; 
  while  ( strncmp(line,"num_nodes ",10)!=0 ) {
    out << line << endl ; 
    in.getline( line, sizeof(line) ) ; 
  }

  sprintf(line+10,"%3d",N-1);
  out << line << endl ; 
  in.getline( line, sizeof(line) ) ; 

  /*            
  out.seekp ( -96, ios::cur ) ; out << setw(3) << N-1 ; 
  out.seekp (  92, ios::cur ) ; out << endl ;      
  */

  // Print unk_node parameters 

  // Get unk_node line ***required***

  while  ( strncmp(line,"unk_node ",9)!=0 ) {
    out << line << endl ; 
    in.getline( line, sizeof(line) ) ; 
  }

  sprintf(line+11,"%9.5f",mod.prior[0]);
  out << line << endl ; 
  in.getline( line, sizeof(line) ) ; 

  // look for unk_node_ss line  ***optional***
  while  ( strncmp(line,"unk_node_ss ",12)!=0 ) {
    if ( strncmp(line,"node ",5)==0 ) break ;
    if ( strncmp(line,"unk_node_rama ",14)==0 ) break ;
    out << line << endl ; 
    in.getline( line, sizeof(line) ) ; 
  }
  // write out new unk_node_ss line
  if  ( strncmp(line,"unk_node_ss ",12)==0 ) {
    // modifications to unk_node_ss line 
    i = 0;  // unknown node
    l_ptr = line+15;
    x = 0.; jj=MDSSP-1 ;    // write out d-vector and get maximum
    for (j=0 ; j<MDSSP ; j++) {
      if (mod.d[i][j] > x) {x = mod.d[i][j]; jj = j ; }
    }
    for (j=0 ; j<MDSSP ; j++) {
      sprintf(l_ptr,"%8.3f",mod.d[i][j] ) ;  // NOTE: eol gets overwritten!
      l_ptr += 8;
    }
    line[12] = dsspletter[jj] ;
    out << line << endl ; 
    in.getline( line, sizeof(line) ) ; 
  }

  // look for unk_node_rama line
  while  ( strncmp(line,"unk_node_rama ",14)!=0 ) {
    if ( strncmp(line,"node ",5)==0 ) break ;
    out << line << endl ; 
    in.getline( line, sizeof(line) ) ; 
  }
  // write out new unk_node_rama line
  if  ( strncmp(line,"unk_node_rama ",14)==0 ) {
    // add modifications to unk_node_rama line here************
    i = 0;  // unknown node
    l_ptr = line+17;
    x = 0.; jj=MRAMA-1 ;    // write out r-vector and get maximum
    for (j=0 ; j<MRAMA ; j++) {
      if (mod.r[i][j] > x) {x = mod.r[i][j]; jj = j; }
      sprintf(l_ptr,"%8.3f",mod.r[i][j]) ;  // NOTE: eol gets overwritten!
      l_ptr += 8;
    }
    line[14] = ramach[jj] ;
    out << line << endl ; 
    in.getline( line, sizeof(line) ) ; 
  }

  // Print known node parameters 

  for ( i=1 ; i<N ; i++ ){

    // look for node line i
    while  ( strncmp(line,"node ",5)!=0 ) {
      if (strncmp(line,"transit_freqs ",14)==0) break ;
      out << line << endl ;  // output miscelaneous lines
      in.getline( line, sizeof(line) ) ; 
    }
    // write out new node 
    if ( strncmp(line,"node ",5)==0 ) {
       // write out new mod.prior, overwrite the \0 character
        sprintf(line+15,"%7.5f",mod.prior[i]); *(line+22) = ' ' ;
#ifdef USE_NEW_RAMA_MAP
        x = 0.; jj=-1 ;
        for (j=0;j<MRAMA;j++) if (mod.r[i][j] > x) {x = mod.r[i][j]; jj=j; }
        if (jj >= 0) {
          sprintf(line+22,"%8.2f",ramacen[jj][0]);
          sprintf(line+30,"%8.2f",ramacen[jj][1]);
          sprintf(line+38,"%8.2f", jj==MRAMA-1 ? 0. : 180. );
          *(line+46) = ' ';
        }
#endif
        out << line << endl ; 
        if ( i==spnode ) strcpy ( spline, line ) ; 

       // write out new b[] vector
        in.getline( line, sizeof(line) )  ;   // 1st line of b[]
        for (j=0 ; j<10 ; j++) sprintf(line+(j*8),"%8.5f",mod.b[i][j]) ;
        out << line << endl ; 
        in.getline( line, sizeof(line) )  ;   // 2nd line of b[]
        for (j=10 ; j<20 ; j++) sprintf(line+((j-10)*8),"%8.5f",mod.b[i][j]) ;
        out << line << endl << flush ; 
        in.getline( line, sizeof(line) )  ;   // whatever follows it
    }

    // look for NODE_SS line
    while  ( strncmp(line,"node_ss ",8)!=0 ) {
      if  ( strncmp(line,"node ",5)==0 ) break;
      if  ( strncmp(line,"transit_freqs ",14)==0 ) break;
      out << line << endl ; 
      in.getline( line, sizeof(line) ) ; 
    }
    if  ( strncmp(line,"node ",5)==0 ) continue ;
    if  ( strncmp(line,"transit_freqs ",14)==0 ) break;

    // output new NODE_SS line
    if ( strncmp(line,"node_ss ",8)==0 ) {
      l_ptr = line+18;
      x = 0.; jj=MDSSP-1 ;    // write out d-vector and get maximum
      for (j=0 ; j<MDSSP ; j++) {
        if (mod.d[i][j] > x) {x = mod.d[i][j]; jj = j ; }
      }
      for (j=0 ; j<MDSSP ; j++) {
        sprintf(l_ptr,"%8.3f",mod.d[i][j]) ;  // NOTE: eol gets overwritten!
        l_ptr += 8;
      }
      line[15] = dsspletter[jj] ;
      out << line << endl ;
      if ( i == spnode ) strcpy ( spline_ss, line );
      in.getline( line, sizeof(line) ) ; // whatever follows node_ss line
    }
    
    // look for NODE_RAMA line
    while  ( strncmp(line,"node_rama ",10)!=0 ) {
      if  ( strncmp(line,"node ",5)==0 ) break;
      if  ( strncmp(line,"transit_freqs ",14)==0 ) break;
      out << line << endl ; 
      in.getline( line, sizeof(line) ) ; 
    }
    if  ( strncmp(line,"node ",5)==0 ) continue ;
    if  ( strncmp(line,"transit_freqs ",14)==0 ) break;

    // output new NODE_RAMA line
    if  ( strncmp(line,"node_rama ",10)==0 ) {
      l_ptr = line+20;
      x = 0.; jj=MRAMA-1 ;    // write out r-vector and get maximum
      for (j=0 ; j<MRAMA ; j++) {
        if (mod.r[i][j] > x) {x = mod.r[i][j]; jj = j; }
        sprintf(l_ptr,"%8.3f",mod.r[i][j]) ;  // NOTE: eol gets overwritten!
        l_ptr += 8;
      }
      line[17] = ramach[jj] ;
      out << line << endl << flush ;
      if ( i == spnode ) strcpy ( spline_rama, line );
      in.getline( line, sizeof(line) ) ; // whatever follows node_rama line
    }

#ifdef USE_CONTEXT
    // look for NODE_CONTEXT line
    while  ( strncmp(line,"node_context ",13)!=0 ) {
      if  ( strncmp(line,"node ",5)==0 ) break;
      if  ( strncmp(line,"transit_freqs ",14)==0 ) break;
      out << line << endl ; 
      in.getline( line, sizeof(line) ) ; 
    }
    if  ( strncmp(line,"node ",5)==0 ) continue ;
    if  ( strncmp(line,"transit_freqs ",14)==0 ) break;

    // output new NODE_CONTEXT line
    if  ( strncmp(line,"node_context ",13)==0 ) {
      l_ptr = line+23;
      x = 0.; jj=MCTXT-1 ;    // write out r-vector and get maximum
      for (j=0 ; j<MCTXT ; j++) {
        if (mod.c[i][j] > x) {x = mod.c[i][j]; jj = j; }
        sprintf(l_ptr,"%8.3f",mod.c[i][j]) ;  // NOTE: eol gets overwritten!
        l_ptr += 8;
      }
      line[20] = context_ch[jj] ;
      out << line << endl << flush ;
      if ( i == spnode ) strcpy ( spline_context, line );
      in.getline( line, sizeof(line) ) ; // whatever follows node_rama line
    }
#endif

  }

  // Output L unit **optional**

  if ( spnode==-1 ){
    for ( i=N-8 ; i<N ; i++ ){

        strcpy(line,"node ") ; 
	sprintf( line+5, "%3d", i ) ; 
	sprintf( line+8, "%s", " z   0") ;
	sprintf( line+14, "%8.5f", mod.prior[i] ) ; 
	sprintf( line+22, "%8.2f", mod.dih[i][0] )  ;
	sprintf( line+30, "%8.2f", mod.dih[i][1] )  ;
	sprintf( line+38, "%8.2f", mod.dih[i][2] )  ;
	sprintf( line+46, " %s", "component=z  99 class=Z ss=" ) ; 
	sprintf( line+74, "%c", mod.ss[i] ) ; 
	out << line << endl ;  

       // write out new b[] vector for split node
        strcpy(line," ") ; // empty line
        for (j=0 ; j<10 ; j++) sprintf(line+(j*8),"%8.5f",mod.b[i][j]) ;
        out << line << endl ; 
        for (j=10 ; j<20 ; j++) sprintf(line+((j-10)*8),"%8.5f",mod.b[i][j]) ;
        out << line << endl << flush ; 
    }
  }


  // If new node by splitting: write parameters for new node 

      if ( spnode> 0 ) {

	cout << "Writing new node information" << endl; 
	cout << "Nodestring: " << spline << endl ; 

       // write out split node line
        strcpy(line,"node ") ; 
	sprintf( line+5, "%3d", N-1 ) ; 
	out << line ; 
        strcpy ( line, spline+8 ) ;
        out << line << endl ; 

       // write out new b[] vector for split node
        strcpy(line," ") ; // empty line
        for (j=0 ; j<10 ; j++) sprintf(line+(j*8),"%8.5f",mod.b[i][j]) ;
        out << line << endl ; 
        for (j=10 ; j<20 ; j++) sprintf(line+((j-10)*8),"%8.5f",mod.b[i][j]) ;
        out << line << endl << flush ; 

       // write node_ss line
        strcpy(line,"node_ss ") ; 
	sprintf( line+8, "%3d", N-1 ) ; 
	out << line ; 
        strcpy ( line, spline_ss+11 ) ;
        out << line << endl ; 

       // write node_rama line 
        strcpy(line,"node_rama ") ; 
	sprintf( line+10, "%3d", N-1 ) ; 
	out << line ; 
        strcpy ( line, spline_rama+13 ) ;
        out << line << endl ; 

       // write node_context line 
        strcpy(line,"node_context ") ; 
	sprintf( line+13, "%3d", N-1 ) ; 
	out << line ; 
        strcpy ( line, spline_context+16 ) ;
        out << line << endl ; 

        in.getline( line, sizeof(line) )  ;   // whatever follows it

      }

  // Write a_ij 

  out << "transit_freqs  " << setw(3) << N << " " 
      << setw(3) << N << " _abcdetc." << endl ;
  for ( i=0 ; i<N ; i++ ){ 
    for ( j=0 ; j<N ; j++ ){
      out << " " << setprecision(5) << setiosflags(ios::fixed ) 
          <<  mod.a[i][j];
      }
    out << endl << flush ;
  }

  in.close()  ; 
  out.close() ; 
 
}
//
//                                May 13, 1998
// 
//   intrans_get.C - nonzero transitions into all nodes 
//                   intrans[i][0] - number of nonzero transitions into i 
//                   intrans[i][j] - node numbers for nonzero transitions
//
//

void intrans_get ( int ***intrans, struct model md ) 
{
  
  int N, i, j, count ; 

  N = md.N ; 

  *intrans = new int*[N] ;
  for ( j = 0 ; j < N ; j++ ){ 
    count = 0 ;
    for ( i = 0 ; i < N ; i++ ){
      if ( md.a[i][j] != 0 ) count++ ; 
    }
    (*intrans)[j] = new int[count+1] ;
    count = 0 ; 
    for ( i = 0 ; i < N ; i++ ){
      if ( md.a[i][j] != 0 ){
	count++ ;
	(*intrans)[j][count] = i ;
      }
      (*intrans)[j][0] = count ; 
    }
  }

}
//
//                                May 13, 1998
// 
//   outrans_get.C - nonzero transitions out of nodes 
//                   outtrans[i][0] - number of nonzero transitions out of i 
//                   outrans[i][j] - node numbers for nonzero transitions
//
//

void outtrans_get ( int ***outtrans, struct model md ) 
{
  
  int N, i, j, count ; 

  N = md.N ;

  *outtrans = new int*[N] ;

  for ( i = 0 ; i < N ; i++ ){
    count = 0 ; 
    for ( j = 0 ; j < N ; j++ ){
      if ( md.a[i][j] != 0 ) count++ ;
    }
    (*outtrans)[i] = new int[count+1] ;    
    count = 0 ;
    for ( j = 0 ; j < N ; j++ ){
      if ( md.a[i][j] != 0 ){
	count++ ;
	(*outtrans)[i][count] = j ;
      }
      (*outtrans)[i][0] = count ; 
    }
  }

}

//
//   				October 30, 1997
//
//     Olambda.C - probability of sequence of residues given model 
//            Divided by: probability of sequence of residues
//                     given background emsission probability 
//
//            Input:  Residue sequence, sequence length, model
//            Output: Probability of residue sequence w. above normalization
// 

extern int NOT_STATE_FLAG; 

float Olambda( char *O_ptr, short int TT, struct model modll )
{
  int i, j, t, istart;
  int O[Tmax];
  float ss; 
  float *alfo, *alfn ; 

// Convert sequence to integer array 

  aminotoi ( O_ptr, TT, O ); 

// Compute probability 

  alfo = new float[modll.N] ; 
  alfn = new float[modll.N] ; 
  
  if ( NOT_STATE_FLAG == 0 ){ istart = 0; } else { istart = 1; } ;   

  if ( NOT_STATE_FLAG == 0 ) { 
    for ( i = 0 ; i < modll.N ; i++ ){ 
      alfo[i] = modll.prior[i] * modll.b[i][O[0]]/modll.bground[O[0]] ;
    }             // Initialize by pi = modll.prior[i]
  } else if ( NOT_STATE_FLAG == 1 ) {
    for ( i = 1 ; i < modll.N ; i++ ){ 
      alfo[i] = modll.a[0][i] * modll.b[i][O[0]]/modll.bground[O[0]] ;
    }             // Initialize by pi = a[0][i]
  }
  
  for ( t = 1; t < TT; t++ ){

    for ( j = istart; j < modll.N; j++ ){
      ss = 0.;
      for ( i = istart; i < modll.N; i++ ){
	ss += alfo[i] * modll.a[i][j] ;	
	if ( NOT_STATE_FLAG == 1 ) ss += alfo[i]  * modll.a[i][0] * modll.a[0][j]; 
      }
      alfn[j] = ss * modll.b[j][O[t]] / modll.bground[O[t]];
    }
    for ( j = istart; j < modll.N; j++ ) alfo[j] = alfn[j];

  }

  ss = 0.;
  for ( i = istart; i < modll.N ; i++ ) ss += alfn[i];

  delete alfo ; delete alfn ;

  return ss ; 

}

//
// features_print - Outputs to stream summary information on model 
//

void features_print( struct model md, int** intrans, int** outtrans, ofstream* stream )
{

  int i, j, N, trans_count, one_count, node_count, sink_count, source_count; 

  N = md.N ; 

  // Number of actual nodes ( nonzero priors )

  node_count = 0 ; 
  for ( i=0 ; i<N ; i++ ){
    if ( md.prior[i] != 0 ) node_count++ ; 
  }

  // Number of non-zero transitions
  
  trans_count = 0 ;
  one_count = 0 ;
  for ( i=0 ; i<N ; i++ ){
    for ( j=0 ; j<N ; j++ ){ 
      if ( md.a[i][j] > 0 ){
	trans_count++ ;
	if ( md.a[i][j] == 1 ) one_count++ ; 
      }
    }
  }
  
  // Find sources
  source_count = 0 ; 
  for ( i=0 ; i<N ; i++ ){
    if ( md.prior[i] != 0 ){
      if ( intrans[i][0] == 0 ){
	//	*stream << "Node " << i << " has no incoming transition" ;
	source_count++ ; 
      }
    }
  }

  // Find sinks
  sink_count = 0 ; 
  for ( i=0 ; i<N ; i++ ){
    if ( md.prior[i] != 0 ){
      if ( outtrans[i][0] == 0 ){
	sink_count++ ; 
	//	*stream << "Node " << i << " has no outgoing transition" ;
      }
    }
  }

  *stream << "Nodes: " << node_count << 
    ", Transitions : " << trans_count << 
    ", Non-unit transitions: " << trans_count-one_count ; 
  if ( source_count ) *stream << ", Sources: " << source_count ; 
  if ( sink_count   ) *stream << ", Sinks: "   << sink_count ; 
  *stream << endl ; 

}

//
//     halflife.c - Persistence length 
//                  For node q, returns length of longest chain 
//                  of nodes starting from q for which product of 
//                  transition probabilites 
//                  is greater than DECAY_FAC 
//
// CB
#define DECAY_FAC 0.1 
int halflife(int q, float w, int **outtrans, struct model mod)
{
  float x ;
  int j,i,maxlife,life ;

  maxlife = 0; 
  life = 0;
  for (i = 1; i <= outtrans[q][0] ; i++) {
    j = outtrans[q][i];
    x = w * mod.a[q][j] ;
    if (x > DECAY_FAC && j != 0) {
      life = 1 + halflife(j,x,outtrans,mod) ;
      if (life > maxlife) {
        maxlife = life ;
      }
    }
  }
  return(maxlife);
}
    


/***************************************************************************

  rel_entropy.c - relative entropy between two aa distributions, log base 2

****************************************************************************/

float rel_entropy ( float* a, float *b ){

  int i;
  float sum=0.;
  float log2=log(2.);

  /*** Assumes that b[i] != 0 ( e.g. our background frequency ) ***/
  for ( i=0 ; i<M ; i++ ){
    if ( a[i] != 0 ){
      sum += a[i] * log ( a[i]/b[i] );
    }
  }
  sum /= log2;

  return sum;

}

/*********************************************************************************/

int O[Tmax];
double brat, rrat, drat, ss;

//
//   				Sep 27, 1998 
//
//     getalpha.C - Compute alpha, using b, r, c, d vectors  
//                       uses FLAG to deteremine which emissions to use
//

void getalpha ( 
	       double **alpha, 
	       struct parent pa, 
	       struct model md, 
	       int **itr ,
	       double *ct ,
               int iflag )
{
  int N, T, i, j, t, ind ;
  T = pa.T;
  N = md.N ;
  aminotoi ( pa.O_ptr, T, O );

  // t = 0 

  ct[0] = 0. ;
  drat = 1.0; rrat = 1.0 ;

  for ( i = 0 ; i < N ; i++ ){ 

    alpha[0][i] = md.prior[i] * flagrat_brdc_fast ( iflag, i, 0, O, pa, md ) ;
    ct[0] += alpha[0][i];

  }

  ct[0] = 1./ct[0] ; // == 1 / sum of alpha[0]
  for ( j = 0; j < md.N; j++ ) alpha[0][j] *= ct[0] ;


  for ( t = 1; t < T; t++ ){

    ct[t] = 0.;  // init scale factor
    for ( j = 0; j < md.N; j++ ){

      ss = 0.;
      for ( i = 0; i < itr[j][0] ; i++ ){
	ind = itr[j][i+1] ; 
	ss += alpha[t-1][ind] * md.a[ind][j] ;
      }

      alpha[t][j] = ss *  flagrat_brdc_fast ( iflag, j, t, O, pa, md ) ;
      ct[t] += alpha[t][j] ;  // inverse of scale factor for next cycle
    }
    if ( ct[t] == 0 ){
      cout << "Error: no incoming transition found for position " << t+1 <<
	" in family " << pa.code << pa.chainID << pa.domainID << endl ; 
      cout << "Position " << t << " Residue: " << pa.O_ptr[t-1] <<
	" DSSP symbol: " << pa.O_ss[t-1] << 
	" Rama symbol: " << pa.O_rama[t-1] <<
#ifdef USE_CONTEXT 
	" Context: " << pa.O_context[t-1] << 
#endif
	endl ;
      cout << "Position " << t+1 << " Residue: " << pa.O_ptr[t] <<
	" DSSP symbol: " << pa.O_ss[t] << 
	" Rama symbol: " << pa.O_rama[t] <<
#ifdef USE_CONTEXT 
	" Context: " << pa.O_context[t] << 
#endif
	endl ;


      exit(8) ;
    }
    ct[t] = 1./ct[t] ;
    for ( j = 0; j < md.N; j++ ) alpha[t][j] *= ct[t] ;

  }
}
//
//   				  sep 27, 1998 
//
//     getbeta.C - Compute beta using bprof
//

void getbeta ( 
               double **beta, 
	       struct parent pa, 
	       struct model md,
	       int **otr ,
               double *ct,
               int iflag  )
{
  int N, T, i, j, t, jnd ;

  T = pa.T;
  N = md.N ;
  aminotoi ( pa.O_ptr, T, O ); 

  for ( i = 0 ; i < md.N ; i++ ) beta[T-1][i] = ct[T-1] ; 

  for ( t = T-2 ; t >= 0 ; t-- ){

    for ( i = 0; i < md.N; i++ ){
      ss = 0.;
      for ( j = 0; j < otr[i][0] ; j++ ) {

	jnd = otr[i][j+1] ;

	ss += md.a[i][jnd] * beta[t+1][jnd] * 
              flagrat_brdc_fast(iflag,jnd,t+1,O,pa,md) ;
      }
      beta[t][i] = ss * ct[t] ;
    }

  }
}


//
//   				         Feb 18, 1999
//
//     getalpha_ns.C - Compute alpha, using b, r, c, d vectors  
//                       uses FLAG to deteremine which emissions to use
//                       Use bprof() in scoring b.
//
//                       Uses terminal and initial distributions 
//                       and NOT-state
//                                               
void getalpha_ns ( 
	       double **alpha, 
	       struct parent pa, 
	       struct model md, 
	       int **itr ,
	       double *ct ,
               int iflag )
{
  int N, T, i, j, t, ind ;
  T = pa.T;
  N = md.N ;
  aminotoi ( pa.O_ptr, T, O );

  // t = 0 

  drat = 1.0; rrat = 1.0 ;

  // ========================== alpha for states @ 0
  ct[0] = 0.;
  for ( j = 1 ; j < N ; j++ ){ 
    alpha[0][j] = md.a[0][j] * flagrat_brdc_fast ( iflag, j, 0, O, pa, md ) ;
    ct[0] += alpha[0][j];
  }
  ct[0] = 1./ct[0] ; // sum of alpha[0]
  for ( j = 1; j < N; j++ ) alpha[0][j] *= ct[0] ;
  // ========================== alpha for NOT-state @ 0
  j = 0;  
  ss = 0.;
  for ( i = 0; i < itr[j][0] ; i++ ){   
    ind = itr[j][i+1] ;
    ss += alpha[0][ind] * md.a[ind][j] ;
  }
  alpha[0][j] = ss ;
  for ( t = 1; t < T; t++ ){
    // ========================== alpha for states @ t
    ct[t] = 0.;  // init scale factor
    for ( j = 1; j < N; j++ ){
      ss = 0.;
      for ( i = 0; i < itr[j][0] ; i++ ){   
        ind = itr[j][i+1] ;
	ss += alpha[t-1][ind] * md.a[ind][j] ;
      }
      alpha[t][j] = ss *  flagrat_brdc_fast ( iflag, j, t, O, pa, md ) ;
      ct[t] += alpha[t][j] ;  // inverse of scale factor for next cycle
    }
    // ========================== error if sum(alpha) == 0.
    if ( ct[t] == 0 ){
      cout << "Error: no incoming transition found for position " << t+1 <<
	" in family " << pa.code << pa.chainID << pa.domainID << endl ; 
      cout << "Position " << t << " Residue: " << pa.O_ptr[t-1] <<
	" DSSP symbol: " << pa.O_ss[t-1] << 
	" Rama symbol: " << pa.O_rama[t-1] <<
#ifdef USE_CONTEXT 
	" Context: " << pa.O_context[t-1] << 
#endif
	endl ;
      cout << "Position " << t+1 << " Residue: " << pa.O_ptr[t] <<
             " DSSP symbol: " << pa.O_ss[t] << 
	     " Rama symbol: " << pa.O_rama[t] <<
#ifdef USE_CONTEXT 
	     " Context: " << pa.O_context[t] << 
#endif
	     endl ;
      for (i = t; i < T; i++) ct[i] = 1.; // allow beta calc 
      return ;  // find other end of bad section

      //exit(8) ;
    } else {  
      ct[t] = 1./ct[t] ;
      for ( j = 1; j < N; j++ ) alpha[t][j] *= ct[t] ;
      // ========================== alpha for NOT-state @ t
      j = 0;  
      ss = 0.;
      for ( i = 0; i < itr[j][0] ; i++ ){   
        ind = itr[j][i+1] ;
        ss += alpha[t][ind] * md.a[ind][j] ;
      }
      alpha[t][j] = ss ;
    }

  }
}
//
//   				  Feb 18, 1999 
//
//     getbeta_ns.C - Compute beta using bprof
//                    NOT-state version 
//
void getbeta_ns ( 
               double **beta, 
	       struct parent pa, 
	       struct model md,
	       int **otr ,
               double *ct,
               int iflag  )
{
  int N, T, i, j, t, jnd ;
  float cct;

  T = pa.T;
  N = md.N ;
  aminotoi ( pa.O_ptr, T, O ); 

  // ========================== beta for state i @ T
  // for ( i = 1 ; i < N ; i++ ) beta[T-1][i] = md.a[i][0] * ct[T-1] ; 
  for ( i = 1 ; i < N ; i++ ) beta[T-1][i] = ct[T-1] ; 
  // ========================== beta for NOT-state @ T
  // i = 0;
  ss = 0.;
  for ( j = 0; j < otr[0][0] ; j++ ) {
    jnd = otr[0][j+1] ;
    ss += md.a[0][jnd] * beta[T-1][jnd] *
	  flagrat_brdc_fast(iflag,jnd,T-1,O,pa,md) ;
  }
  beta[T-1][0] = ss ;   // do not scale beta for NOT-state

  for ( t = T-2 ; t >= 0 ; t-- ){
    cct = 0.;
    // ========================== beta for state i @ t
    for ( i = 1; i < N; i++ ){
      ss = 0.;
      for ( j = 0; j < otr[i][0] ; j++ ) {
        jnd = otr[i][j+1] ;
	ss += md.a[i][jnd] * beta[t+1][jnd] * 
	  flagrat_brdc_fast(iflag,jnd,t+1,O,pa,md) ;
      }
      beta[t][i] = ss * ct[t] ;
      cct += beta[t][i] ;
    }
    // ========================== beta for NOT-state  @ t
    // i = 0;   
    ss = 0.;
    for ( j = 0; j < otr[0][0] ; j++ ) {
      jnd = otr[0][j+1] ;
      ss += md.a[0][jnd] * beta[t][jnd] *
	  flagrat_brdc_fast(iflag,jnd,t,O,pa,md) ;
    }
    beta[t][0] = ss ;  // do not scale beta for NOT-state
    if ( cct == 0 ){
      cout << "Error: no outgoing transition found for position " << t+1 <<
	      " in family " << pa.code << pa.chainID << pa.domainID << endl ; 
      cout << "Position " << t << " Residue: " << pa.O_ptr[t-1] <<
	      " DSSP symbol: " << pa.O_ss[t-1] << 
	      " Rama symbol: " << pa.O_rama[t-1] <<
#ifdef USE_CONTEXT 
	      " Context: " << pa.O_context[t-1] << 
#endif
	      endl ;
      cout << "Position " << t+1 << " Residue: " << pa.O_ptr[t] <<
	      " DSSP symbol: " << pa.O_ss[t] << 
	      " Rama symbol: " << pa.O_rama[t] <<
#ifdef USE_CONTEXT 
	      " Context: " << pa.O_context[t] << 
#endif
	      endl ;
      exit(8) ;
    }

  }
}


//				
//				September 30, 1997
//
//      aminotoi.C - converts array of residues to array of integers
//
//                   Input: O_ptr  Pointer to residue sequence 
//                          N      Length of sequence 
//
void aminotoi ( char* O_ptr, short int N, int a[] ){ 

  int oneamino ( char );

  int i;
  
  for( i = 0; i < N ; i++ ) a[i] = oneamino( *(O_ptr+i) ) ;

}

int oneamino( char res )
{
  int  i;
  char aa_alf[22] = "ACDEFGHIKLMNPQRSTVWYX"; // X : aa is unknown
   
  for ( i=0 ; i<21 ; i++ ){
    if ( res == aa_alf[i] ) return i;
  }

  cerr << "Amino acid " << res <<" not in alphabet" << endl;
  exit(8);

}
