//  Isites_Covar.C
// This program predicts I-sites based on the I-sites-2 Library
// Covariance models for local structure motifs.
// Mon Nov 10 22:27:12 EST 2003
//!!*=========================================================================
//!!** Copyright 2003-2004 Chris Bystroff
//!!**
//!!** Permission is hereby granted to use, execute, copy, distribute, modify,
//!!** and distribute in modified form this software for non-profit purposes,
//!!** provided this notice is retained in any and all copies.
//!!**
//!!** For for-profit usage, please contact bystrc@rpi.edu
//!!*=========================================================================
#include <fstream.h>
#include <iostream.h>
#include <iomanip.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "vall_tools.C"
#define NCEN 9
#define NCLUST 192
#define SSSIZE 19
#include "readisl.C"
#include "getmda.C"

//void readisiteslibrary(char *isiteslibrary, struct cluster **ISL, int *nclust);
//float getmda(int L,float *phi,float *psi,float *omg, parent pr, int ires);


main (int argc, char *argv[]) 
{
   int *NNa;
   int i,j,k,ii,jj, na, iclust, lcf, icen, jcen, ntop ;
   float qscore, cscore, sscore, alpha=0.5, cf, y = 0., yy;
   float x = 0., mda, ww, b0, b1;
   float pi=3.1415927 ;
   float pa[20];

   struct parent *dbase;
   struct cluster *ISL;
   int nclust, nt,nn;
   float *topcf, *topmda;
   int *isite;
   //   char libfile[80];
   float W[19][20];
   char res1[20], ssstr[SSSIZE];
   float ***** tensor;
   
   if (argc < 3) {
     cout << "Usage: xisites_covar vall isl Ntop [Cluster]" << endl;
     cout << "vall = drct format sequence profile file.  " << endl ;
     cout << "isl = ISITES2 library format, including COVARTENSOR" << endl ;
     cout << "Ntop = number of top scoring I-sites to output for each position." << endl;
     cout << "Cluster = (optional) input a cluster number to use. Default is all clusters." << endl;
     cout << "isites_covar.C  v.22-FEB-2005 " << endl;
     exit(1) ;
   }
   res1[0]='A'; res1[1]='C'; res1[2]='D'; res1[3]='E';
   res1[4]='F'; res1[5]='G'; res1[6]='H'; res1[7]='I';
   res1[8]='K'; res1[9]='L'; res1[10]='M'; res1[11]='N';
   res1[12]='P'; res1[13]='Q'; res1[14]='R'; res1[15]='S';
   res1[16]='T'; res1[17]='V'; res1[18]='W'; res1[19]='Y';
   ssstr[SSSIZE] = '\0';


   NNa = new int;
   dbasef(argv[1], &dbase, NNa) ;
   cout << "REMARK predicting I-sites in file " << argv[1] << *NNa << " sequences." << endl << flush ;


   // temporary---------------
   // Sum the background AA probs from the database.
   for (k=0;k<20;k++) pa[k] = 0.0;
   for (na=0; na<*NNa; na++) {
      for (j=0;j<dbase[na].T;j++) {
        for (k=0;k<20;k++) pa[k] += dbase[na].O_profile[j][k]; 
      }
   } 
   for (k=0; k<20; k++) x += pa[k];
   for (k=0; k<20; k++) pa[k] /= x ;    
   // temporary---------------
    cout << "TRAININGSET " << argv[1] << endl;
   // temporary---------------
    
   // Read new library
   ISL = new cluster [NCLUST] ;
   tensor = new float **** [NCLUST] ;
   readisl(argv[2], ISL, tensor, pa, &nclust);
    cout << "BACKGROUND AA PROBS " ;
    for (k=0; k<20; k++) {
      cout << setw(8) << setprecision(5) << setiosflags(ios::fixed);
      cout << pa[k];
    }
    cout << endl;
   if (nclust > NCLUST) {
     cout << "INCREASE MEMORY ALLOCATION for NCLUST" << endl;
     exit(1);
   } else {
     cout << "REMARK  " << nclust << " I-sites clusters read. " << endl ;
   }
   // Initialize Wi using LIBRARY14.0
   // readisiteslibrary(argv[2], &ISL, &nclust);
   // cout << "REMARK Done Reading Isites Library14.0" << endl;
   // overwrite ww if input on the commandline 
   iclust = 0;
   int myclust = 0;
   if (argc > 3) {
     ntop = atoi(argv[3]) ;
   } else {
     ntop = 1;
   }
   cout << "REMARK Getting " << ntop << " top scores for all I-sites in" << argv[2] << endl;

   if (argc > 4) {
     myclust = atoi(argv[4]) ;
     cout << "REMARK Using only cluster " << myclust << endl;
     ntop = 1;
   } else {
     myclust = 0;
     cout << "REMARK Using all clusters "  << endl;
   }

   isite = new int [ntop];
   topcf = new float [ntop];
   topmda = new float [ntop];

   ww = -1.0;
   // if (argc > 4) {
   //   ww = ISL[iclust].ww = atof(argv[4]) ;
   //   cout << "REMARK Fixed covariance weight=" << ISL[iclust].ww << endl;
   // } else {
   //   cout << "REMARK Covariance weights will be read from the ISL file." << endl;
   //   ww = -1.0;
   // } 

   // Initialize the true-false array to "-1" = neither t nor f

   // for (na=0; na<*NNa; na++) {
     // cout << dbase[na].T << " " << flush ;
     // dbase[na].tf = new short int [dbase[na].T];
     // dbase[na].O_context = new char [dbase[na].T];
     // dbase[na].bestcf = new float [dbase[na].T];
     // dbase[na].qscore = new float [dbase[na].T];
     // dbase[na].cscore = new float [dbase[na].T];
     // for (i=0; i<dbase[na].T ; i++) {
     //   dbase[na].tf[i] = -1;
     //   dbase[na].bestcf[i] = 0;
     //   dbase[na].O_context[i] = '0';
     // }
   // }

   // cout << "REMARK Done initializing tf " << endl ;
   // for (iclust=0;iclust<nclust;iclust++) {
   //   cout << ISL[iclust].ID << " " << ISL[iclust].mdacut << " " << ISL[iclust].dmecut << endl;
   // }
     // diagnostic: do one cluster
     // move this to where the covar score is calculated
     // if (myclust != 0 && ISL[iclust].ID != myclust) continue;
     // read ww, b0, b1
     // read tensor here
     //
     for (na=0; na<*NNa; na++) {
       for (i=0; i<dbase[na].T ; i++) {
          dbase[na].cen[i] = getcentroid(dbase[na].O_profile[i],pa) ;
          // cout << "Centroid : " << na << " " << i << " " << dbase[na].cen[i] << endl;
       }
     }
     for (na=0; na<*NNa; na++) {
       // cout << dbase[na].code << endl ; 
       cout << "SEQUENCE: " << setw(6) << na << " " ;
       for (ii=0;ii<4;ii++) cout << dbase[na].code[ii] ;
       cout << " length=" << setw(5) << dbase[na].T << " keychar=%";
       cout << endl;
//  3    4  103 1abrB  104 INPK                EbHH                  4021   0.357  159.5 1bia_  201 EEBE                 false
cout << "  111 2222 3333 44444 5555 666666666666666666 777777777777777777 888888 9999999 000000 11111 2222 3333333333333333333 44444" << endl;
cout << "   Ns   Nc Ires codeI  res sequence           sec_struct         clustr confdnc   mda  pcode pres psec_struct         t/f" << endl;

       for (i=0; i<dbase[na].T ; i++) {
               
         for (k=0;k<ntop;k++) { topcf[k]=0.; isite[k]=0;}

         for (iclust=0;iclust<nclust;iclust++) {
       
           if (myclust != 0 && ISL[iclust].ID != myclust) continue;

           if (i > dbase[na].T - ISL[iclust].Lwi ) continue;

           // qscore = sum of log likelihoods (using pseudocounts)
           // calculate score using multiple sequences
           qscore = 0;
           for (j=0; j<ISL[iclust].Lwi; j++) {
             for (k=0; k<20; k++) {
                     W[j][k] = log((dbase[na].O_profile[i+j][k] + alpha*pa[k])/(pa[k]*(1+alpha)));
                     qscore += W[j][k]*ISL[iclust].Wi[j][k] ;
             }
           }
           // sscore is single-sequence score
           // sscore = 0;
           // for (j=0; j<ISL[iclust].Lwi; j++) {
           //   for (k=0; k<20; k++) {
           //     if ( dbase[na].O_ptr[i+j] == res1[k]) break;
           //   }
           //   sscore += ISL[iclust].Wi[j][k] ;
           // }
             // calculate confidence
           // step through confidence curve
           // lcf = 0;
           // while (qscore < ISL[iclust].cfscore[lcf]) {
           //   lcf++;
           //   if (lcf >= ISL[iclust].Lcf) break;
           // }
           // if (lcf == 0) {
           //   cf = ISL[iclust].cf[lcf] ;
           // } else {  // interpolate
           //   x = (ISL[iclust].cfscore[lcf-1]-ISL[iclust].cfscore[lcf]);
           //   y = (qscore - ISL[iclust].cfscore[lcf])/x;
           //   cf = ISL[iclust].cf[lcf] + y*(ISL[iclust].cf[lcf-1] - ISL[iclust].cf[lcf]) ;
           // }
           // covariance score qscore += covarscore
           cscore = 0.;
           if (ISL[iclust].ww > 0.) {
             for (ii=ISL[iclust].ohang; ii<ISL[iclust].L+ISL[iclust].ohang; ii++) {
               icen = dbase[na].cen[i+ii];
               for (jj=ii+1; jj<ISL[iclust].L+ISL[iclust].ohang; jj++) {
                 if (ii==jj) continue;
                 jcen = dbase[na].cen[i+jj];
                 cscore += tensor[iclust][ii-ISL[iclust].ohang][jj-ISL[iclust].ohang][icen][jcen];
               }
             }
           }
           // cout << iclust << " " <<  ISL[iclust].ID << " qscore=" << qscore << "  cscore=" << cscore << flush << endl;
  
           mda = getmda(ISL[iclust].L,ISL[iclust].phi,ISL[iclust].psi,ISL[iclust].omg,dbase[na],i+ISL[iclust].ohang);
           // dme = getdme(ISL[iclust].L,ISL[iclust].phi,ISL[iclust].psi,ISL[iclust].omg,dbase[na], i);
           // IGNORE non-linear transformation if ww was set. By default (ww = -1.) use ww from isl file.
           if (ww < 0.) {   // use ww and linearfit
             // apply relative weight to qscore, cscore
             cf = (ISL[iclust].ww)*cscore + (1.-ISL[iclust].ww)*qscore ;
             // tranform to probability   
             cf = (1/pi)*atan(ISL[iclust].b0 + ISL[iclust].b1*cf) + ISL[iclust].b2 ;
             // truncate confidence at 0.99. 
             if (cf >= 1.00) cf = 0.99;
           } else {
             cf = (ww)*cscore + (1.-ww)*qscore ;
           }
           j=ISL[iclust].ohang + i;
           k = ntop-1;
           while (cf > topcf[k-1]) {
             topcf[k] = topcf[k-1];
             topmda[k] = topmda[k-1];
             isite[k] = isite[k-1];
             k--;
             if  (k==0) break;
           }
           if (cf > topcf[k]) {
             topcf[k] = cf;
             topmda[k] = mda;
             isite[k] = iclust;
           }
           // for (j=ISL[iclust].ohang; j<ISL[iclust].L+ISL[iclust].ohang; j++) {
           //   if (mda < ISL[iclust].mdacut) {    /// this is to sum the number of trues possible
           //     dbase[na].O_context[i+j] = '1';
           //   }
           //   if (cf > dbase[na].bestcf[i+j]) {
           //     dbase[na].bestcf[i+j] = cf;
           //     dbase[na].qscore[i+j] = qscore;
           //     dbase[na].cscore[i+j] = cscore;
           //     if (mda < ISL[iclust].mdacut) {
           //       dbase[na].tf[i+j] = ISL[iclust].ID ;
           //       // cout << iclust << " " <<  ISL[iclust].ID << 
           //       //      setw(8) << setprecision(2) << setiosflags(ios::fixed) 
           //       //      << " " << qscore << " " << cscore << " T" << endl ;
           //     } else {
           //       dbase[na].tf[i+j] = 0;
           //       // cout << iclust << " " <<  ISL[iclust].ID << 
           //       //      setw(8) << setprecision(2) << setiosflags(ios::fixed) 
           //       //      <<" " << qscore << " " << sscore << " F" << endl ;
           //     }
           //   }
         }
         //  Write out top ntop isites
         for (k=0; k<ntop; ++k) {
           if (isite[k]==0) continue;
           cout << "% ";
           cout << setw(3) << na
                << setw(5) << k 
                << setw(5) << j 
                << " " ;
           for (ii=0;ii<4;ii++) cout << dbase[na].code[ii] ;
           cout << dbase[na].chainID 
                << setw(5) << dbase[na].resseq[j]
                << " " ;
           // draw sequence string
           for (ii=0; ii<ISL[isite[k]].L; ++ii) {
              ssstr[ii] = dbase[na].O_ptr[j+ii];
           }
           for (ii=ISL[isite[k]].L; ii<SSSIZE;++ii) ssstr[ii] = ' ';
           // cout << setw(15) << ssstr;  // sequence
           cout << ssstr;  // sequence
          
           // draw local structure string for query
           for (ii=0; ii<ISL[isite[k]].L; ++ii) {
              // ssstr[ii] = dbase[na].O_ss[j+ii];
              ssstr[ii] = Ramatype(dbase[na].O_phi[j+ii],dbase[na].O_psi[j+ii],dbase[na].O_omg[j+ii]);
           }
           for (ii=ISL[isite[k]].L; ii<SSSIZE;++ii) ssstr[ii] = ' ';
           cout << ssstr;

           // output prediction confidence and mda (true = (mda < 120.))
           // cout << setw(7) << isite[k]
           cout << setw(6) << ISL[isite[k]].ID << " ";
           cout << setw(7) << setprecision(3) << setiosflags(ios::fixed) << topcf[k] 
                << setw(7) << setprecision(1) << setiosflags(ios::fixed) << topmda[k];
           cout << setw(5) << ISL[isite[k]].pcode << ISL[isite[k]].pchain << setw(5) << ISL[isite[k]].pstart
                << " ";

           // draw predicted local structure string 
           for (ii=0; ii<ISL[isite[k]].L; ++ii) {
              // cout << Ramatype(ISL[isite[k]].phi[ii],ISL[isite[k]].psi[ii],ISL[isite[k]].omg[ii]);
              ssstr[ii] = Ramatype(ISL[isite[k]].phi[ii],ISL[isite[k]].psi[ii],ISL[isite[k]].omg[ii]);
           }
           for (ii=ISL[isite[k]].L; ii<SSSIZE;++ii) ssstr[ii] = ' ';
           cout << ssstr;

           if (topmda[k] <= ISL[isite[k]].mdacut) { cout << " TRUE"; } else { cout << " false"; }
           cout << endl;
         }
                 
       }
     }
     // (add loop over all clusters)
     // cout << "REMARK Done with cluster " << iclust << " " << ISL[iclust].ID << endl ;
   //}
   // Now we have data to sum benchmark
   // int ntot = 0;
   // nt = 0;
   // nn = 0;
   // for (na=0; na<*NNa; na++) {
   //   for (i=0; i<dbase[na].T ; i++) {
   //     if (dbase[na].O_context[i] == '1') ntot++;
   //     if (dbase[na].tf[i] >= 1) nt++;
   //     if (dbase[na].tf[i] >= 0) nn++;
   //     // diagnostic
   //     cout << dbase[na].code << dbase[na].chainID << " " << dbase[na].O_ptr[i] ;
   //     cout << " " << i ;
   //     cout << " " << dbase[na].cen[i] ;
   //     //
   //     cout << setw(8) << setprecision(2) << setiosflags(ios::fixed) 
   //                  << dbase[na].qscore[i] << setw(8) << dbase[na].cscore[i]  
   //                  << setw(8) << dbase[na].bestcf[i] << " "
   //                  << setw(8) << dbase[na].tf[i] << endl ;
   //   }
   // }

   // if (nn > 0) { x = (float) ntot/(float) nn ;} else { x = 0.;}
   // cout << "REMARK Total trues possible =" << ntot << " out of " << nn << " = " 
   //      << setprecision(8) << x << endl;

   // if (nn > 0) { x = (float) nt/(float) nn ;} else { x = 0.;}
   // cout << "REMARK Your benchmark value = "  ;
   // cout << setprecision(8) << x << endl ;

   // if (ntot > 0) { x = (float) nt/(float) ntot ;} else { x = 0.;}
   // cout << "REMARK Fraction of possible correct  = "  ;
   // cout << setprecision(8) << x << endl ;
   
}


//
//float getmda(int L,float *phi,float *psi,float *omg, parent pr, int ires)
//{
//    float dd, mda = 0;
//    int i;
//
//    for (i=0; i < (L-1); i++) {
//      // delta-phi 1 -> L-1
//      dd = fabs(phi[i+1] - pr.O_phi[ires+i+1]);
//      if (dd > 180.) dd = 360. - dd;
//      // cout << " phi " << setw(8) << setprecision(2) << setiosflags(ios::fixed) << phi[i+1] ;
//      // cout << setw(8) << setprecision(2) << pr.O_phi[ires+i+1] << setw(7) << dd ;
//      if (dd > mda) mda = dd;
//
//      // delta-psi 0 -> L-2
//      dd = fabs(psi[i] - pr.O_psi[ires+i]);
//      if (dd > 180.) dd = 360. - dd;
//      // cout << " psi " << setw(8) << setprecision(2) << setiosflags(ios::fixed) << psi[i] ;
//      // cout << setw(8) << setprecision(2) << pr.O_psi[ires+i] <<setw(7) <<  dd ;
//      if (dd > mda) mda = dd;
//
//      // delta-omg 0 -> L-2
//      dd = fabs(omg[i+1] - pr.O_omg[ires+i+1]);
//      if (dd > 180.) dd = 360. - dd;
//      // cout << " omg " << setw(8) << setprecision(2) << setiosflags(ios::fixed) << omg[i+1] ;
//      // cout << setw(8) << setprecision(2) << pr.O_omg[ires+i+1] <<setw(7) <<  dd ;
//      if (dd > mda) mda = dd;
//      // cout << endl; 
//    }
//    // cout << setw(8) << setprecision(2) << mda << endl;
//
//    return(mda);
//}
