#!/usr/bin/perl
# profile2drct.pl: compile the PSI-BLAST profile into the binary drct file.

use strict;

# drct i/o
my $template;
my $RECORDSIZE;
my $record;
my @fields;


###########################
### DRCT data structure ###
###########################
  # 2 2 4 1 1 1 1 12 12 4 80 4 4 1 1 6
  # record size: 136 bytes
  # (0)seq_list (1)resSeq (2)structure_name (3)chainID (4)seq_resid (5)ss (6)rotamer (7-9)xyzca (10-12)ang (13)csum (14-33)vall_pro (34)gap_freq (35)ins_freq (36)domainID (37)ctx (38-40)clusterID
  # $template = "s s A4 A A A A f f f f f f f f f f f f f f f f f f f f f f f f f f f f f A A s3";
  $template = "s s A4 A A A A f3 f3 f f20 f f A A s3";
  $RECORDSIZE =  136;
###########################



### start of drct record structure ##
my $drctLen = 0;
my @drctSeq = ();
my @drctSeq_list = ();
my @drctResSeq = ();
my @drctStructure_name = ();
my @drctChainID = ();
my @drctSeq_resid = ();
my @drctSS = ();
my @drctRotamer = ();
my @drctXYZca = ();
my @drctAng = ();
my @drctCsum = ();
my @drctVall_pro = ();
my @drctGap_freq = ();
my @drctIns_freq = ();
my @drctDomainID = ();
my @drctCtx = ();
my @drctClusterID = ();
### end of drct record structure ##

die ("Usage: profile2drct.pl blast_output blast_profile drct [code=blst]\nNote: code is a 4-letter word.\n") unless scalar(@ARGV) >= 3;

# my ($blastfile, $profile, $drctfile) = @_;
my $blastfile = shift;
my $profile = shift;
my $drctfile = shift;
my $code = shift;
if (length ($code) == 0) {
  $code = "blst";
} else {
  $code =~ s/^(....).*/$1/;
}

# read seqeunce from blats output
my $seq = &blastSeq ($blastfile);
my $seqLen = length ($seq);
print "$seq\n";

# read in sequence profile
&readProfile ($profile, \@drctVall_pro);
# !!!! change the aa order of the profile to be consistant with
# !!!! the old version!
&rearrange (\@drctVall_pro, $seqLen);

my ($i, $j) = 0;
for ($i = 0; $i < $seqLen; $i++) {
  for $j (0 .. 19) {
    printf "%4.2f ", $drctVall_pro[$i][$j];
  }
  print "\n";
}

# prepare drct data
for ($i = 0; $i < $seqLen; $i++) {
  $drctSeq_list[$i] = $i + 1;
  $drctResSeq[$i] = $i + 1;
  $drctStructure_name[$i] = $code;
  $drctChainID[$i] = " ";
  $drctSeq[$i] = substr ($seq, $i, 1);
  $drctSS[$i] = " ";
  $drctRotamer[$i] = " ";
  $drctXYZca[$i][0] = 0;
  $drctXYZca[$i][1] = 0;
  $drctXYZca[$i][2] = 0;
  $drctAng[$i][0] = 0;
  $drctAng[$i][1] = 0;
  $drctAng[$i][2] = 0;
  $drctCsum[$i] = 0;
  $drctGap_freq[$i] = 0;
  $drctIns_freq[$i] = 0;
  $drctDomainID[$i] = " ";
  $drctCtx[$i] = " ";
  $drctClusterID[$i][0] = 0;
  $drctClusterID[$i][1] = 0;
  $drctClusterID[$i][2] = 0;
}

# write to drct file.
$drctLen = $seqLen;
&writeDRCT;

print "profile2drct.pl v.0.1 Yu Shao (shaoy\@rpi.edu) Nov 2003\n";
print "done.\n";

# blastSeq: read and return the query sequence from the psi-blast output.
sub blastSeq {
  my $file = shift;
  my $seq = "";

  open (IN, $file) || die ("$file: file open failed.\n");
  # 1. Read the QUERY line.
  # 2. Stop reading when see the first "Searching.................done"
  #    after the first QUERY is read.
  while (<IN>) {
    chomp;
    if (/^QUERY/) {
      split;
      $seq .= $_[2];
      while (<IN>) {
        chomp;
        if (/Searching\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.\.done/) {
          last;
        }
        if (/^QUERY/) {
          split;
          $seq .= $_[2];
        }
      }
      last;
    }
  }
  close (IN);

  # remove gap characters
  $seq =~ s/-+//g;

  if (length($seq) == 0) {
    printf STDERR "No blast hits!\n";
    print "profile2drct.pl v.0.1 Yu Shao (shaoy\@rpi.edu) Nov 2003\n";
    print "done.\n";
    exit;
  }

  return $seq;
}


# readProfile: read the profile data from the profile generated
#              by psi-blast.
sub readProfile {
  my $profile = shift;
  my $p = shift;

  my ($i, $j) = 0;

  open (IN, $profile) || die ("$profile: file open failed.\n");
  while (<IN>) {
    if (/  A  R  N  D/) {
      while (<IN>) {
        if (/^$/ || /Lambda/) {
          last;
        }
        chomp;
        split;
        for $j (0 .. 19) {
          $p->[$i][$j] = $_[$j + 22] / 100;
        }
        $i++;
      }
      last;
    }
  }
  close (IN);

  if ($i != length ($seq)) {
    print "ERROR\n";
  }
}

sub rearrange {

  my $prof = shift;
  my $seqLen = shift;
  my @newarray = ();
  my ($i, $j) = 0;
  my (%t1, %t2) = ();

#        0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16 17 18 19 
# blast: A  R  N  D  C  Q  E  G  H  I  L  K  M  F  P  S  T  W  Y  V
# drct : A  C  D  E  F  G  H  I  K  L  M  N  P  Q  R  S  T  V  W  Y
  $t1{0} = "A";   $t2{"A"} = 0;
  $t1{1} = "R";   $t2{"R"} = 14;
  $t1{2} = "N";   $t2{"N"} = 11;
  $t1{3} = "D";   $t2{"D"} = 2;
  $t1{4} = "C";   $t2{"C"} = 1;
  $t1{5} = "Q";   $t2{"Q"} = 13;
  $t1{6} = "E";   $t2{"E"} = 3;
  $t1{7} = "G";   $t2{"G"} = 5;
  $t1{8} = "H";   $t2{"H"} = 6;
  $t1{9} = "I";   $t2{"I"} = 7;
  $t1{10} = "L";  $t2{"L"} = 9;
  $t1{11} = "K";  $t2{"K"} = 8;
  $t1{12} = "M";  $t2{"M"} = 10;
  $t1{13} = "F";  $t2{"F"} = 4;
  $t1{14} = "P";  $t2{"P"} = 12;
  $t1{15} = "S";  $t2{"S"} = 15;
  $t1{16} = "T";  $t2{"T"} = 16;
  $t1{17} = "W";  $t2{"W"} = 18;
  $t1{18} = "Y";  $t2{"Y"} = 19;
  $t1{19} = "V";  $t2{"V"} = 17;

  for $i (0 .. $seqLen) {
    for $j (0 .. 19) {
      $newarray[$i][$t2{$t1{$j}}] = $prof->[$i][$j];
    }
  }

  for $i (0 .. $seqLen) {
    for $j (0 .. 19) {
      $prof->[$i][$j] = $newarray[$i][$j];
    }
  }
}

sub writeDRCT {
  
  my $i;
  my $j;
  my @entry = ();

  open (DRCT, ">$drctfile") || die ("$drctfile: output file open failed.\n");
  binmode (DRCT);
  
  for ($i = 0; $i < $drctLen; $i++) {
    @entry = ($drctSeq_list[$i], $drctResSeq[$i], $drctStructure_name[$i], $drctChainID[$i], $drctSeq[$i], $drctSS[$i], $drctRotamer[$i]);
    for ($j = 0; $j < 3; $j++) {
      @entry = (@entry, $drctXYZca[$i][$j]);
    }
    for ($j = 0; $j < 3; $j++) {
      @entry = (@entry, $drctAng[$i][$j]);
    }
    @entry = (@entry, $drctCsum[$i]);
    for ($j = 0; $j < 20; $j++) {
      @entry = (@entry, $drctVall_pro[$i][$j]);
    }
    @entry = (@entry, $drctGap_freq[$i], $drctIns_freq[$i], $drctDomainID[$i], $drctCtx[$i]);
    for ($j = 0; $j < 3; $j++) {
      @entry = (@entry, $drctClusterID[$i][$j]);
    }

    $record = pack ($template, @entry);
    print (DRCT $record);

#   print ("$#entry  ");
#   for ($j = 0; $j <= $#entry; $j++) {
#     print ("\"$entry[$j]\"  ");
#   }
#   print "\n";
  }

  close (DRCT);

}
