
// DSSP symbols HEGST_ 
// Three state reduction :
// Reverted to old convention ( consistent with Jones 1999 ) on April 7, 1999
// Old convention H->H E->E {GST_}->L
// "New" convention (consistent with Rost etc. ) {H,G}->H E->E {ST_}->L
// Maps below that are consistent with old convention:
// threestate (modified 9/17/99)
// dcompress (modified 4/7/99)
// inv_dcompress (created 9/17/98)

#include <stdlib.h>
#include <math.h>
//#include "hmm.h"
#include "ss_maps.h"

#define USE_NEW_RAMA_MAP

float ramacen[MRAMA][2]  = {
  {  -61.91 ,  -45.20 },   // H
  { -109.78 ,   20.88 },   // G
  {  -70.58 ,  147.22 },   // B
  { -132.89 ,  142.43 },   // E
  { -135.03 ,   77.26 },   // d
  {  -85.03 ,   72.26 },   // b
  { -165.00 ,  175.00 },   // e
  {   55.88 ,   38.62 },   // L
  {   85.82 ,   -0.03 },   // l
  {   80.00 , -170.00 },   // x
  {  -70.00 ,  150.00 }    // c=cis!!
} ;


/***************************************************************************
 inv_dcompress.c - inversion of dcompress dssp symbol order 
                   all L are assigned to T
****************************************************************************/

void inv_dcompress ( struct model *modl ) 
{
  int i ; 
  
  modl->dground[0] = modl->dground[0] ;
  modl->dground[1] = modl->dground[1] ;
  modl->dground[4] = modl->dground[2] ; 
  modl->dground[2] = 0. ;
  modl->dground[3] = 0. ; 
  modl->dground[5] = 0. ;

  for ( i=0 ; i<modl->N ; i++ ){ 
    modl->d[i][0] = modl->d[i][0] ;
    modl->d[i][1] = modl->d[i][1] ;
    modl->d[i][4] = modl->d[i][2] ;
    modl->d[i][2] = 0. ; 
    modl->d[i][3] = 0. ;
    modl->d[i][5] = 0. ;    
  }

}

/****************************************************************************
  dcompress.c -  convert mod.d distribution to HEL 
****************************************************************************/

void dcompress ( struct model *modl ) 
{
  int i ; 

  modl->dground[0] = modl->dground[0];
  modl->dground[1] = modl->dground[1];
  modl->dground[2] = modl->dground[2]+modl->dground[3] + modl->dground[4] + modl->dground[5] ; 
  modl->dground[3] = 0. ; 
  modl->dground[4] = 0. ;
  modl->dground[5] = 0. ;

  for ( i=0 ; i<modl->N ; i++ ){ 
    modl->d[i][0] = modl->d[i][0] + modl->d[i][2] ;
    modl->d[i][1] = modl->d[i][1] ;
    modl->d[i][2] = modl->d[i][3] + modl->d[i][4] + modl->d[i][5] ; 
    modl->d[i][3] = 0. ; 
    modl->d[i][4] = 0. ;
    modl->d[i][5] = 0. ;    
  }

}

int irama ( char rama_symbol ){

  int i;
#ifdef USE_NEW_RAMA_MAP
  char rama_string[MRAMA+1] = "HGBEdbeLlxc" ;
#else
  char rama_string[MRAMA+1] = "HELX" ;
#endif

  for ( i=0 ; i<MRAMA ; i++ ){
    if ( rama_symbol == rama_string[i] ) return i;
  }
  return -1 ;
  
}

/***************************************************************************/

// 4-7-99 Converted to Jones etc convention

int threestate ( char dsspsymbol ) {

  if ( dsspsymbol == 'H' || dsspsymbol == 'I' )
    return 0 ; 
  if ( dsspsymbol == 'E' ) 
    return 1 ; 
  if ( dsspsymbol == 'S' || dsspsymbol == 'T' || dsspsymbol == 'G' || \
       dsspsymbol == 'B' || dsspsymbol == '_' || dsspsymbol == 'L' )
    /*  April 2000 : added last entry to accomodate predefined HEL */
    return 2 ; 
 
  return -1 ; 
}

/***************************************************************************/

int icontext ( char context_symbol ){

  int i; 
  char context_string[MCTXT+1] = "amnhdstul_" ;

  for ( i=0 ; i<MCTXT ; i++ ){
    if ( context_symbol == context_string[i] ) return i ;
  }
  return -1 ;

}


/***************************************************************************/

#ifdef USE_NEW_RAMA_MAP
char  Ramatype(float ph, float ps, float om)
{
  char rama_string[MRAMA+1] = "HGBEdbeLlxc" ;
  float d,dmin,ds,df;
  int i, icen;
  extern float ramacen[11][2];

  if (om < 90. && om > -90.) return 'c';

  dmin = 999.;
  for (i=0;i<MRAMA-1;i++) {
    df = fabs(ph - ramacen[i][0]); if (df>180.) df = 360.-df;
    ds = fabs(ps - ramacen[i][1]); if (ds>180.) ds = 360.-ds;
    d = sqrt(ds*ds + df*df);
    if (d < dmin) { icen=i; dmin=d; }
  }
  return rama_string[icen] ;
}
#else
char  Ramatype(float ph, float ps, float om)
{

  if (ph == 0.) return('?') ; // first residue
  if (ps == 0.) return('?') ; // last residue
  if (om < 90. && om > -90.) return('?') ; // cis peptide, set to unknown

    if (ph > 0. ) {
      if (ps > -90. && ps < 90.) {
        return('L') ; // left-handed helix
      } else {
        if (ph > 30. && ph < 150.) {
          return('X') ; //  rare
        } else {
          return('E') ; // fringe extended
        }
      }
    } else {
      if (ps < 40. && ps > -120.) {
        return('H') ; // helix
      } else {
        return('E') ; // extended
      }
    }
}
#endif






