#include <iostream>
#include <iomanip>
#include <fstream> 
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <string>
#include "hmm.h"
#include "vall_tools.h"
#include "ss_maps.C"
#include "fasta_tools.h"
#include "centroids.h"

#define MAXfamily 3000
//#define Vmax 50000  // Temporary storage size in dbasef (no. of positions )
#define Vmax 300000
#define TOOSMALL 20
#define PROFFLAG 1
#define NEWDBASE 1

//
// smooth_strand - transforms context symbols in database such that 
//                 n becomes m if n has an m as a nearest neighbor
//
void smooth_strand ( struct parent **dbase, int *NNa )
{
  int na, t, T, i, n_count, m_count, check_count=0, old_diff; 
  struct parent *mydb; 
  char *ctxt_temp ; // working context character array 

  mydb = *dbase ; 
  
  n_count = 0 ; 
  m_count = 0; 
  for ( na=0 ; na< *NNa ; na++ ){
    T = mydb[na].T ; 
    for ( t=0 ; t<T ; t++ ){
      if ( mydb[na].O_context[t] == 'm' ) m_count++ ;  
      if ( mydb[na].O_context[t] == 'n' ) n_count++ ;  
    }
  }
  old_diff = n_count - m_count ; 
  cout << "Database has " << m_count << " m positions and " << n_count << " n positions" << endl ; 

  cout << "In database: converting n to m if n has nearest neighbor m" << endl ; 

  for ( na=0 ; na< *NNa ; na++ ){

    T = mydb[na].T ; 
    ctxt_temp = new char [T] ; 

    // t=0 
    if ( mydb[na].O_context[0] == 'n' && mydb[na].O_context[1] == 'm' ){
      ctxt_temp[0] = 'm' ;
      check_count++ ; 
    } else 
      ctxt_temp[0] = mydb[na].O_context[0] ; 
    // 0 < t < T-1  
    for ( t=1 ; t<T-1 ; t++ ){
      if ( mydb[na].O_context[t] == 'n' && 
	   ( mydb[na].O_context[t-1] == 'm' || mydb[na].O_context[t+1] == 'm' ) ){
	ctxt_temp[t] = 'm' ;
	check_count++ ; 
      } else 
	ctxt_temp[t] = mydb[na].O_context[t] ;
    }
    // t=T-1
    if ( mydb[na].O_context[T-1] == 'n' && mydb[na].O_context[T-2] == 'm' ){
      ctxt_temp[T-1] = 'm' ;
      check_count++ ; 
    } else 
      ctxt_temp[T-1] = mydb[na].O_context[T-1] ;

    // copy working context vector to database and reset
    
    for ( t=0 ; t<T ; t++ ){
      mydb[na].O_context[t] = ctxt_temp[t] ; 
    }
    delete ctxt_temp ; 

  }// end loop over parents


  n_count = 0 ; 
  m_count = 0; 
  for ( na=0 ; na< *NNa ; na++ ){
    T = mydb[na].T ; 
    for ( t=0 ; t<T ; t++ ){
      if ( mydb[na].O_context[t] == 'm' ) m_count++ ;  
      if ( mydb[na].O_context[t] == 'n' ) n_count++ ;  
    }
  }
  cout << "Database has " << m_count << " m positions and " << n_count << " n positions" << endl ; 
  if ( old_diff - (n_count-m_count) != 2 * check_count ){
    cout << "Error: Number of n to m conversions inconsistent " << endl ;
    cout << "Number of conversions in loops : " << check_count << endl ; 
    exit(8) ; 
  }

}

//
//  low_filter - set observed frequences in profiles to 0
//               if they are below a given threshold  
//
#define THRESHOLD 0.1

void low_filter ( struct parent **dbase, int *NNa )
{
  int na, t, T, i, ncount=0, Tsum=0, subcount; 
  float renorm; 
  struct parent *mydb;

  mydb = *dbase ; 

  cout << "Filtering out low count observations" << endl; 
  for ( na=0 ; na< *NNa ; na++ ){
    T = mydb[na].T ; 
    Tsum += T ;
    for ( t=0 ; t<T ; t++ ){
      renorm = 1.;
      subcount = 0 ; 
      for ( i=0 ; i<M ; i++ ){
	if ( mydb[na].O_profile[t][i] < THRESHOLD && 
	     mydb[na].O_profile[t][i] > 0. ){
	  renorm -= mydb[na].O_profile[t][i];
	  mydb[na].O_profile[t][i] = 0.;
	  //	  cout << "removed: parent " << na << " position " << t <<
	  //   " aa " << i << endl;
	  subcount++ ;
	  ncount++ ;
	}
	if ( subcount == 20 ){
	  cout << "Error: all counts removed in parent " << na <<
	    " position " << t << endl ;
	  exit(8);
	}
      }
      // cout << "na " << na << " t " << t << " " << renorm << endl ;
      for ( i=0 ; i<M ; i++ ){
	mydb[na].O_profile[t][i] /= renorm;
      }
    }
  }
  cout << "Number of profile entries filtered out, per position: " <<
    float(ncount)/float(Tsum) << endl ;

}
// 
//                                    February 5, 1998
//
//      dbasef.C - Reads database, returns protein sequences, dihedral angles, 
//                structure codes 
//


void dbasef ( char * vallfile, struct parent **dbase, int * NNa ) 
{

  int i, na, count, T , ilen;
  short int oldseq_list ;
  char code[MAXfamily][4];
  char chainID[MAXfamily];
  int seqlength[MAXfamily];
  short int *centemp;
  short int *resseqtemp;
  char *Otemp, *sstemp ;
  float **prof ;
  float *phi, *psi, *omg ;
  float *xx, *yy, *zz ;
  float *gap_freq_temp, *ins_freq_temp ;
  char domain[MAXfamily];
  char *context;
  
  struct parent *mydb;

  struct dbaserec rec;
  cout << "beginning..." << endl << flush ;

  // allocate temporary storage
  Otemp = new char[Vmax];
  sstemp = new char[Vmax];
  phi = new float[Vmax];
  psi = new float[Vmax];
  omg = new float[Vmax];
  xx = new float[Vmax];
  yy = new float[Vmax];
  zz = new float[Vmax];
  if ( PROFFLAG ) {
    prof = new float*[Vmax] ;
    for (i=0;i<Vmax;i++) prof[i] = new float[M] ;
  }
  // if ( NEWDBASE ){
    gap_freq_temp = new float[Vmax] ;
    ins_freq_temp = new float[Vmax] ;
    context       = new char[Vmax] ;
    centemp = new short int[Vmax];
    resseqtemp = new short int[Vmax];
  // }

  //
  // open vallfile.. binary file of sequence + structure data.
  //
  cerr << "Opening " << vallfile << endl << flush ;

  ifstream in ( vallfile, ios::in ) ;
  if ( in.bad() ) {
    cerr << "Unable to open database file: " << vallfile << "\n";
    exit (8);
  }

  //
  // Read whole file into temporary buffer 
  //

  count = 0 ; 
  na = 0;
  // do { in.read( (unsigned char *) &rec, sizeof( rec  )  ); } 
  do { in.read( (char *) &rec, sizeof( rec  )  ); } 
    while ( rec.seq_resid == '-' || rec.seq_resid == 'X' ) ;    
  while (in.gcount() == sizeof(rec)) {
    // start a new protein chain
    for (i=0;i<4;i++) code[na][i] = rec.code[i] ; 
    chainID[na] = rec.chainID ;
    if ( NEWDBASE ) {
      domain[na] = rec.domainID;
    }
    ilen = 0;
    oldseq_list = -9999 ; 
    while ( rec.resSeq > oldseq_list ){
      if ( rec.seq_resid == 'Z' ) rec.seq_resid = 'Q' ;  
      if ( rec.seq_resid == 'B' ) rec.seq_resid = 'N' ; 
      Otemp[count]  = rec.seq_resid ;
      phi[count]    = rec.dih[0] ; 
      psi[count]    = rec.dih[1] ; 
      omg[count]    = rec.dih[2] ; 
      xx[count]    = rec.xyzca[0] ; 
      yy[count]    = rec.xyzca[1] ; 
      zz[count]    = rec.xyzca[2] ; 
      centemp[count]    = rec.centroid ; 
      resseqtemp[count]    = rec.resSeq ; 
      sstemp[count] = rec.ss ;
      // diagnostic 
      // cout << rec.seq_resid << " " << rec.resSeq << " "<< rec.centroid << endl;
      if ( PROFFLAG ) { 
	      for (i = 0 ; i < M; i++){
	        prof[count][i]    = rec.profile[i] ;
	      }
      }
      if ( NEWDBASE ) {
	      gap_freq_temp[count] = rec.gap_freq;
	      ins_freq_temp[count] = rec.ins_freq;
	      context[count]       = rec.context;
      }
      ++count ;                        // no. excluding deletions
      if (count >= Vmax) {
        cout << "Exceeded maximum allocation!!! increase Vmax." << endl ;
        exit(1);
      }
      ++ilen;                          // length of chain 
      //      oldseq_list = rec.seq_list ;     // no. including deletions
      oldseq_list = abs(rec.resSeq) ;     // no. ungapped sequences
      // read in the next record
      // do { in.read( (unsigned char *) &rec, sizeof( rec  )  ); } 
      do { in.read( (char *) &rec, sizeof( rec  )  ); } 
        while ( rec.seq_resid == '-' || rec.seq_resid == 'X' ) ;    
      if (rec.seq_list == 1) break ;
    }  
    seqlength[na] = ilen;
    // Output commented out .  Wed Jun 19 08:18:51 EDT 2002
    // cerr << na << code[na] << " " << chainID[na] 
     //     << " Length=" << ilen << endl ;
    if (ilen < TOOSMALL) {
       //  cerr << "Chain too short! " 
       //       << code[na] << " " << chainID[na] 
       //       << " Length=" << ilen << endl << flush ;
      count -= ilen ;
    } else { na++; }
      // cerr << endl ;
  }  
  // cerr << " all done... na=" << na << endl << flush ;
  *NNa = na;
  // cerr << " all done...NNa=" <<  *NNa << endl << flush ;
  in.close() ; 

  // 
  // Allocate dbase and transfer everything.
  // 

  // *dbase = new parent[*NNa];
  *dbase = new parent[na];
  cout << "new " << flush ;
  mydb = *dbase;
  cout << "Allocating space for "<< *NNa << " chains." << endl;
  cout << "Total number of residues= "<< count <<  endl;
  count = 0;
  for (na=0 ; na < *NNa; na++) {
    // cout << "starting " << flush ;
    T = seqlength[na] ; 
    mydb[na].T     = T ;
    for (i=0;i<4;i++) mydb[na].code[i] = code[na][i] ; 
    mydb[na].chainID = chainID[na] ; 
    if ( NEWDBASE ) {
      mydb[na].domainID = domain[na] ;
    }
    // allocate sub-arrays
    mydb[na].O_ptr = new char[T] ;
    mydb[na].O_phi = new float[T] ;
    mydb[na].O_psi = new float[T] ;
    mydb[na].O_omg = new float[T] ;
    mydb[na].x = new float[T] ;
    mydb[na].y = new float[T] ;
    mydb[na].z = new float[T] ;
    mydb[na].O_ss  = new char[T] ;
    mydb[na].O_rama  = new char[T] ;
    // cout << "..profile " << flush ;
    if ( PROFFLAG ) {
      mydb[na].O_profile  = new float*[T] ;
      for (i = 0; i < T; i++) mydb[na].O_profile[i] = new float[M]; 
    }
    if ( NEWDBASE ) {
      mydb[na].O_gap      = new float[T];
      mydb[na].O_ins      = new float[T];
      mydb[na].O_context  = new char[T];
      mydb[na].cen  = new short int[T];
      mydb[na].resseq  = new short int[T];
    }
    // cout << "..substring " << flush ;
    //transfer from temporary memory
    substring  ( Otemp+count, mydb[na].O_ptr, T ) ; 
    substring2 ( phi + count, mydb[na].O_phi, T ) ; 
    substring2 ( psi + count, mydb[na].O_psi, T ) ; 
    substring2 ( omg + count, mydb[na].O_omg, T ) ; 
    substring2 ( xx + count, mydb[na].x, T ) ; 
    substring2 ( yy + count, mydb[na].y, T ) ; 
    substring2 ( zz + count, mydb[na].z, T ) ; 
    substring  ( sstemp+count,mydb[na].O_ss , T ) ; 

    // convert angles to Rama character: HELX
    // for ( i = 0 ; i < T ; i++ ) *(mydb[na].O_rama+i) = 
    //   Ramatype(mydb[na].O_phi[i],mydb[na].O_psi[i],mydb[na].O_omg[i]) ;

    // save profile data
    if ( PROFFLAG ){
      for (i = 0; i < T; i++){
	      substring2 ( prof[i+count], mydb[na].O_profile[i], 20 ) ; 
      }
    }

    if ( NEWDBASE ){
      substring2 ( gap_freq_temp + count, mydb[na].O_gap, T );
      substring2 ( ins_freq_temp + count, mydb[na].O_ins, T );
      substring  ( context       + count, mydb[na].O_context, T );
      substring3 ( centemp+count,mydb[na].cen , T ) ; 
      substring3 ( resseqtemp+count,mydb[na].resseq , T ) ; 
    }
    // cout << na << endl ; 
    count += seqlength[na];
  }
  cout << "Total number of residues stored in mem= "<< count << endl ;
  cout << "Last chain "<< setw(6) << na << " : " ;
  for (i=0;i<4;i++) cout << mydb[na-1].code[i] ;
  cout << endl ;

  // Clean up temporary storage

  delete []Otemp; Otemp=NULL;
  delete []sstemp; sstemp=NULL;
  delete []phi; phi=NULL;
  delete []psi; psi=NULL;
  delete []omg; omg=NULL;
  delete []xx; xx=NULL;
  delete []yy; yy=NULL;
  delete []zz; zz=NULL;
  delete []centemp; centemp=NULL;
  delete []resseqtemp; resseqtemp=NULL;
  if ( PROFFLAG ) { 
    for (i=0;i<Vmax;i++) delete[] prof[i];
    delete []prof; prof=NULL;
  }

}

/*************************************************************************
       Substring of string  of length N 
**********************************************************************/

void substring ( char *string, char *substr_ptr, short int N )
{
  int i ; 
  
  for ( i = 0 ; i < N ; i++ ) *(substr_ptr + i) = *(string + i) ;
 
}

void substring2 ( float *string, float *substr_ptr, short int N )
{

  int i ; 

  for ( i = 0 ; i < N ; i++ ) *(substr_ptr + i) = *(string + i) ;
 
}

void substring3 ( short int *string, short int *substr_ptr, short int N )
{

  int i ; 

  for ( i = 0 ; i < N ; i++ ) *(substr_ptr + i) = *(string + i) ;
 
}


// 
//                                    april 18, 2001
//
//      dbaseput.C - Write database
//


void dbaseput ( char * vallfile, char * outfile, struct parent *dbase, int * NNa ) 
{

  int i, na, count, T , ilen;
  short int oldseq_list ;
  // int seqlength[MAXfamily];
  
  struct dbaserec rec;
  cout << "beginning..." << endl << flush ;

  //
  // open vallfile.. binary file of sequence + structure data.
  //
  cout << "opening" << vallfile << endl << flush ;

  ifstream in ( vallfile, ios::in ) ;

  if ( !in ) {
    cerr << "Unable to open input file: " << vallfile << "\n";
    exit (8);
  }

  ofstream output ( outfile, ios::out | ios::binary ) ;

  if ( !output ) {
    cerr << "Unable to open output file: " << outfile << "\n";
    exit (8);
  }

  //
  // Read whole file into temporary buffer 
  //

  count = 0 ; 
  na = 0;
  // do { in.read( (unsigned char *) &rec, sizeof( rec  )  ); } 
  do { in.read( (char *) &rec, sizeof( rec  )  ); } 
    while ( rec.seq_resid == '-' || rec.seq_resid == 'X' ) ;    
  while (in.gcount() == sizeof(rec)) {
    // start a new protein chain
    ilen = 0;
    oldseq_list = -9999 ; 
    while ( rec.resSeq > oldseq_list ){
      if ( rec.seq_resid == 'Z' ) rec.seq_resid = 'Q' ;  
      if ( rec.seq_resid == 'B' ) rec.seq_resid = 'N' ; 
      ++count ;                        // no. excluding deletions
      rec.centroid = dbase[na].cen[ilen] ;
      ++ilen;                          // length of chain 
      //      oldseq_list = rec.seq_list ;     // no. including deletions
      oldseq_list = abs(rec.resSeq) ;     // no. ungapped sequences
      // write record with centroid
      cout << na << " " << rec.seq_resid << " " << rec.centroid << endl;
      output.write((char *) &rec, sizeof( rec  ) ) ;
      // read in the next record
      // do { in.read( (unsigned char *) &rec, sizeof( rec  )  ); } 
      do { in.read( (char *) &rec, sizeof( rec  )  ); } 
        while ( rec.seq_resid == '-' || rec.seq_resid == 'X' ) ;    
      if (rec.seq_list == 1) break ;
    }  
    // seqlength[na] = ilen;
    // cout << na << code[na] << " " << chainID[na] 
    //      << " Length=" << ilen << flush ;
    if (ilen < TOOSMALL) {
      cout << "Chain too short! " ;
    //           << code[na] << " " << chainID[na] 
    //           << " Length=" << ilen << endl << flush ;
      count -= ilen ;
    } else { na++; }
  }  
  // *NNa = na;
  in.close() ; 
  output.close() ; 
  cout << " all done..." << *NNa << " " << na << endl << flush ;

  // 
  // Allocate dbase and transfer everything.
  // 

  // Clean up temporary storage

}

int getcentroid(float * profile, float * pa) 
{
    int Lcen=9, icen=0, i,k;
    float mindd=-9999999., alpha=ALPH_1;
    float dd, qscore;

    for (i=0;i<Lcen;++i) {
      qscore = 0;
      for (k=0; k<20; k++) {
         dd = log((profile[k] + alpha*pa[k])/(pa[k]*(1+alpha)));
         qscore += dd*centroid[i][k] ;
      }
      // diagnostic
      // cout << "Center " << i << " distance
      if (qscore > mindd) {
        icen = i;
        mindd = qscore;
      }
    }

    return icen;
}
